import asyncio
import gc
import os
import ssl
import socket
import sys
import time
import uuid

from docker import APIClient

import aiomysql
import pymysql
import pytest


PY_35 = sys.version_info >= (3, 5)
if PY_35:
    import uvloop
else:
    uvloop = None


@pytest.fixture(scope='session')
def unused_port():
    def f():
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(('127.0.0.1', 0))
            return s.getsockname()[1]
    return f


def pytest_generate_tests(metafunc):
    if 'loop_type' in metafunc.fixturenames:
        loop_type = ['asyncio', 'uvloop'] if uvloop else ['asyncio']
        metafunc.parametrize("loop_type", loop_type)

    # if 'mysql_tag' in metafunc.fixturenames:
    #     tags = set(metafunc.config.option.mysql_tag)
    #     if not tags:
    #         tags = ['5.7']
    #     elif 'all' in tags:
    #         tags = ['5.6', '5.7', '8.0']
    #     else:
    #         tags = list(tags)
    #     metafunc.parametrize("mysql_tag", tags, scope='session')


# This is here unless someone fixes the generate_tests bit
@pytest.yield_fixture(scope='session')
def mysql_tag():
    return '5.6'


@pytest.yield_fixture
def loop(request, loop_type):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(None)

    if uvloop and loop_type == 'uvloop':
        loop = uvloop.new_event_loop()
    else:
        loop = asyncio.new_event_loop()

    yield loop

    if not loop._closed:
        loop.call_soon(loop.stop)
        loop.run_forever()
        loop.close()
    gc.collect()
    asyncio.set_event_loop(None)


@pytest.mark.tryfirst
def pytest_pycollect_makeitem(collector, name, obj):
    if collector.funcnamefilter(name):
        if not callable(obj):
            return
        item = pytest.Function(name, parent=collector)
        if 'run_loop' in item.keywords:
            return list(collector._genfunctions(name, obj))


@pytest.mark.tryfirst
def pytest_pyfunc_call(pyfuncitem):
    """
    Run asyncio marked test functions in an event loop instead of a normal
    function call.
    """
    if 'run_loop' in pyfuncitem.keywords:
        funcargs = pyfuncitem.funcargs
        loop = funcargs['loop']
        testargs = {arg: funcargs[arg]
                    for arg in pyfuncitem._fixtureinfo.argnames}
        loop.run_until_complete(pyfuncitem.obj(**testargs))
        return True


def pytest_runtest_setup(item):
    if 'run_loop' in item.keywords and 'loop' not in item.fixturenames:
        # inject an event loop fixture for all async tests
        item.fixturenames.append('loop')


def pytest_ignore_collect(path, config):
    if 'pep492' in str(path):
        if sys.version_info < (3, 5, 0):
            return True


def pytest_addoption(parser):
    parser.addoption("--mysql_tag", action="append", default=[],
                     help=("MySQL server versions. "
                           "May be used several times. "
                           "Available values: 5.6, 5.7, 8.0, all"))
    parser.addoption("--no-pull", action="store_true", default=False,
                     help="Don't perform docker images pulling")


@pytest.fixture
def mysql_params():
    params = {"host": os.environ.get('MYSQL_HOST', 'localhost'),
              "port": int(os.environ.get('MYSQL_PORT', 3306)),
              "user": os.environ.get('MYSQL_USER', 'root'),
              "db": os.environ.get('MYSQL_DB', 'test_pymysql'),
              "password": os.environ.get('MYSQL_PASSWORD', ''),
              "local_infile": True,
              "use_unicode": True,
              }
    return params


# TODO: fix this workaround
@asyncio.coroutine
def _cursor_wrapper(conn):
    cur = yield from conn.cursor()
    return cur


@pytest.yield_fixture
def cursor(connection, loop):
    cur = loop.run_until_complete(_cursor_wrapper(connection))
    yield cur
    loop.run_until_complete(cur.close())


@pytest.yield_fixture
def connection(mysql_params, loop):
    coro = aiomysql.connect(loop=loop, **mysql_params)
    conn = loop.run_until_complete(coro)
    yield conn
    loop.run_until_complete(conn.ensure_closed())


@pytest.yield_fixture
def connection_creator(mysql_params, loop):
    connections = []

    @asyncio.coroutine
    def f(**kw):
        conn_kw = mysql_params.copy()
        conn_kw.update(kw)
        _loop = conn_kw.pop('loop', loop)
        conn = yield from aiomysql.connect(loop=_loop, **conn_kw)
        connections.append(conn)
        return conn

    yield f

    for conn in connections:
        loop.run_until_complete(conn.ensure_closed())


@pytest.yield_fixture
def pool_creator(mysql_params, loop):
    pools = []

    @asyncio.coroutine
    def f(**kw):
        conn_kw = mysql_params.copy()
        conn_kw.update(kw)
        _loop = conn_kw.pop('loop', loop)
        pool = yield from aiomysql.create_pool(loop=_loop, **conn_kw)
        pools.append(pool)
        return pool

    yield f

    for pool in pools:
        pool.close()
        loop.run_until_complete(pool.wait_closed())


@pytest.yield_fixture
def table_cleanup(loop, connection):
    table_list = []
    cursor = loop.run_until_complete(_cursor_wrapper(connection))

    def _register_table(table_name):
        table_list.append(table_name)

    yield _register_table
    for t in table_list:
        # TODO: probably this is not safe code
        sql = "DROP TABLE IF EXISTS {};".format(t)
        loop.run_until_complete(cursor.execute(sql))


@pytest.fixture(scope='session')
def session_id():
    """Unique session identifier, random string."""
    return str(uuid.uuid4())


@pytest.fixture(scope='session')
def docker():
    return APIClient(version='auto')


@pytest.fixture(scope='session')
def mysql_server(unused_port, docker, session_id, mysql_tag, request):
    if not request.config.option.no_pull:
        docker.pull('mysql:{}'.format(mysql_tag))

    # bound IPs do not work on OSX
    host = "127.0.0.1"
    host_port = unused_port()

    # As TLS is optional, might as well always configure it
    ssl_directory = os.path.join(os.path.dirname(__file__),
                                 'ssl_resources', 'ssl')
    ca_file = os.path.join(ssl_directory, 'ca.pem')
    tls_cnf = os.path.join(os.path.dirname(__file__),
                           'ssl_resources', 'tls.cnf')

    ctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1)
    ctx.check_hostname = False
    ctx.load_verify_locations(cafile=ca_file)
    # ctx.verify_mode = ssl.CERT_NONE

    container_args = dict(
        image='mysql:{}'.format(mysql_tag),
        name='aiomysql-test-server-{}-{}'.format(mysql_tag, session_id),
        ports=[3306],
        detach=True,
        host_config=docker.create_host_config(
            port_bindings={3306: (host, host_port)},
            binds={
                ssl_directory: {'bind': '/etc/mysql/ssl', 'mode': 'ro'},
                tls_cnf: {'bind': '/etc/mysql/conf.d/tls.cnf', 'mode': 'ro'},
            }
        ),
        environment={'MYSQL_ROOT_PASSWORD': 'rootpw'}
    )

    container = docker.create_container(**container_args)

    try:
        docker.start(container=container['Id'])

        # MySQL restarts at least 4 times in the container before its ready
        time.sleep(10)

        server_params = {
            'host': host,
            'port': host_port,
            'user': 'root',
            'password': 'rootpw',
            'ssl': ctx
        }
        delay = 0.001
        for i in range(100):
            try:
                connection = pymysql.connect(
                    db='mysql',
                    charset='utf8mb4',
                    cursorclass=pymysql.cursors.DictCursor,
                    **server_params)

                with connection.cursor() as cursor:
                    cursor.execute("SHOW VARIABLES LIKE '%ssl%';")

                    result = cursor.fetchall()
                    result = {item['Variable_name']:
                              item['Value'] for item in result}

                    assert result['have_ssl'] == "YES", \
                        "SSL Not Enabled on docker'd MySQL"

                    cursor.execute("SHOW STATUS LIKE '%Ssl_version%'")

                    result = cursor.fetchone()
                    # As we connected with TLS, it should start with that :D
                    assert result['Value'].startswith('TLS'), \
                        "Not connected to the database with TLS"

                break
            except Exception as err:
                time.sleep(delay)
                delay *= 2
        else:
            pytest.fail("Cannot start MySQL server")

        container['host'] = host
        container['port'] = host_port
        container['conn_params'] = server_params

        yield container
    finally:
        docker.kill(container=container['Id'])
        docker.remove_container(container['Id'])



import asyncio
from pymysql import NotSupportedError
from aiomysql.cursors import SSCursor
from tests import base
from tests._testutils import run_until_complete

from aiomysql import ProgrammingError, InterfaceError


class TestSSCursor(base.AIOPyMySQLTestCase):
    data = [
        ('America', '', 'America/Jamaica'),
        ('America', '', 'America/Los_Angeles'),
        ('America', '', 'America/Lima'),
        ('America', '', 'America/New_York'),
        ('America', '', 'America/Menominee'),
        ('America', '', 'America/Havana'),
        ('America', '', 'America/El_Salvador'),
        ('America', '', 'America/Costa_Rica'),
        ('America', '', 'America/Denver'),
        ('America', '', 'America/Detroit'), ]

    @asyncio.coroutine
    def _prepare(self, conn):
        cursor = yield from conn.cursor()
        # Create table
        yield from cursor.execute('DROP TABLE IF EXISTS tz_data;')
        yield from cursor.execute('CREATE TABLE tz_data ('
                                  'region VARCHAR(64),'
                                  'zone VARCHAR(64),'
                                  'name VARCHAR(64))')

        yield from cursor.executemany(
            'INSERT INTO tz_data VALUES (%s, %s, %s)', self.data)
        yield from conn.commit()
        yield from cursor.close()

    @run_until_complete
    def test_ssursor(self):
        # affected_rows = 18446744073709551615
        conn = self.connections[0]
        cursor = yield from conn.cursor(SSCursor)
        # Create table
        yield from cursor.execute('DROP TABLE IF EXISTS tz_data;')
        yield from cursor.execute(('CREATE TABLE tz_data ('
                                   'region VARCHAR(64),'
                                   'zone VARCHAR(64),'
                                   'name VARCHAR(64))'))

        # Test INSERT
        for i in self.data:
            yield from cursor.execute(
                'INSERT INTO tz_data VALUES (%s, %s, %s)', i)
            self.assertEqual(conn.affected_rows(), 1,
                             'affected_rows does not match')
        yield from conn.commit()

        # Test update, affected_rows()
        yield from cursor.execute('UPDATE tz_data SET zone = %s', ['Foo'])
        yield from conn.commit()

        self.assertEqual(cursor.rowcount, len(self.data),
                         'Update failed. affected_rows != %s' % (
                         str(len(self.data))))

        yield from cursor.close()
        yield from cursor.close()

    @run_until_complete
    def test_sscursor_fetchall(self):
        conn = self.connections[0]
        cursor = yield from conn.cursor(SSCursor)

        yield from self._prepare(conn)
        yield from cursor.execute('SELECT * FROM tz_data')
        fetched_data = yield from cursor.fetchall()
        self.assertEqual(len(fetched_data), len(self.data),
                         'fetchall failed. Number of rows does not match')

    @run_until_complete
    def test_sscursor_fetchmany(self):
        conn = self.connections[0]
        cursor = yield from conn.cursor(SSCursor)
        yield from self._prepare(conn)
        yield from cursor.execute('SELECT * FROM tz_data')
        fetched_data = yield from cursor.fetchmany(2)
        self.assertEqual(len(fetched_data), 2,
                         'fetchmany failed. Number of rows does not match')

        yield from cursor.close()
        # test default fetchmany size
        cursor = yield from conn.cursor(SSCursor)
        yield from cursor.execute('SELECT * FROM tz_data;')
        fetched_data = yield from cursor.fetchmany()
        self.assertEqual(len(fetched_data), 1)

    @run_until_complete
    def test_sscursor_executemany(self):
        conn = self.connections[0]
        yield from self._prepare(conn)
        cursor = yield from conn.cursor(SSCursor)
        # Test executemany
        yield from cursor.executemany(
            'INSERT INTO tz_data VALUES (%s, %s, %s)', self.data)
        msg = 'executemany failed. cursor.rowcount != %s'
        self.assertEqual(cursor.rowcount, len(self.data),
                         msg % (str(len(self.data))))

    @run_until_complete
    def test_sscursor_scroll_relative(self):
        conn = self.connections[0]
        yield from self._prepare(conn)
        cursor = yield from conn.cursor(SSCursor)
        yield from cursor.execute('SELECT * FROM tz_data;')
        yield from cursor.scroll(1)
        ret = yield from cursor.fetchone()
        self.assertEqual(('America', '', 'America/Los_Angeles'), ret)

    @run_until_complete
    def test_sscursor_scroll_absolute(self):
        conn = self.connections[0]
        yield from self._prepare(conn)
        cursor = yield from conn.cursor(SSCursor)
        yield from cursor.execute('SELECT * FROM tz_data;')
        yield from cursor.scroll(2, mode='absolute')
        ret = yield from cursor.fetchone()
        self.assertEqual(('America', '', 'America/Lima'), ret)

    @run_until_complete
    def test_sscursor_scroll_errors(self):
        conn = self.connections[0]
        cursor = yield from conn.cursor(SSCursor)

        yield from cursor.execute('SELECT * FROM tz_data;')

        with self.assertRaises(NotSupportedError):
            yield from cursor.scroll(-2, mode='relative')

        yield from cursor.scroll(2, mode='absolute')

        with self.assertRaises(NotSupportedError):
            yield from cursor.scroll(1, mode='absolute')
        with self.assertRaises(ProgrammingError):
            yield from cursor.scroll(2, mode='not_valid_mode')

    @run_until_complete
    def test_sscursor_cancel(self):
        conn = self.connections[0]
        cur = yield from conn.cursor(SSCursor)
        # Prepare ALOT of data

        yield from cur.execute('DROP TABLE IF EXISTS long_seq;')
        yield from cur.execute(
            """ CREATE TABLE long_seq (
                  id int(11)
                )
            """)

        ids = [(x) for x in range(100000)]
        yield from cur.executemany('INSERT INTO long_seq VALUES (%s)', ids)

        # Will return several results. All we need at this point
        big_str = "x" * 10000
        yield from cur.execute(
            """SELECT '{}' as id FROM long_seq;
            """.format(big_str))
        first = yield from cur.fetchone()
        self.assertEqual(first, (big_str,))

        @asyncio.coroutine
        def read_cursor():
            while True:
                res = yield from cur.fetchone()
                if res is None:
                    break
        task = self.loop.create_task(read_cursor())
        yield from asyncio.sleep(0, loop=self.loop)
        assert not task.done(), "Test failed to produce needed condition."
        task.cancel()
        try:
            yield from task
        except asyncio.CancelledError:
            pass

        with self.assertRaises(InterfaceError):
            yield from conn.cursor(SSCursor)



# ported from:
# https://github.com/aio-libs/aiopg/blob/master/aiopg/sa/engine.py
import asyncio

import aiomysql
from .connection import SAConnection
from .exc import InvalidRequestError, ArgumentError
from ..utils import _PoolContextManager, _PoolAcquireContextManager
from ..cursors import Cursor


try:
    from sqlalchemy.dialects.mysql.pymysql import MySQLDialect_pymysql
except ImportError:  # pragma: no cover
    raise ImportError('aiomysql.sa requires sqlalchemy')


_dialect = MySQLDialect_pymysql(paramstyle='pyformat')
_dialect.default_paramstyle = 'pyformat'


def create_engine(minsize=1, maxsize=10, loop=None,
                  dialect=_dialect, pool_recycle=-1,  compiled_cache=None,
                  **kwargs):
    """A coroutine for Engine creation.

    Returns Engine instance with embedded connection pool.

    The pool has *minsize* opened connections to PostgreSQL server.
    """
    coro = _create_engine(minsize=minsize, maxsize=maxsize, loop=loop,
                          dialect=dialect, pool_recycle=pool_recycle,
                          compiled_cache=compiled_cache, **kwargs)
    compatible_cursor_classes = [Cursor]
    # Without provided kwarg, default is default cursor from Connection class
    if kwargs.get('cursorclass', Cursor) not in compatible_cursor_classes:
        raise ArgumentError('SQLAlchemy engine does not support '
                            'this cursor class')
    return _EngineContextManager(coro)


async def _create_engine(minsize=1, maxsize=10, loop=None,
                         dialect=_dialect, pool_recycle=-1,
                         compiled_cache=None, **kwargs):

    if loop is None:
        loop = asyncio.get_event_loop()
    pool = await aiomysql.create_pool(minsize=minsize, maxsize=maxsize,
                                      loop=loop,
                                      pool_recycle=pool_recycle, **kwargs)
    conn = await pool.acquire()
    try:
        return Engine(dialect, pool, compiled_cache=compiled_cache, **kwargs)
    finally:
        pool.release(conn)


class Engine:
    """Connects a aiomysql.Pool and
    sqlalchemy.engine.interfaces.Dialect together to provide a
    source of database connectivity and behavior.

    An Engine object is instantiated publicly using the
    create_engine coroutine.
    """

    def __init__(self, dialect, pool, compiled_cache=None, **kwargs):
        self._dialect = dialect
        self._pool = pool
        self._compiled_cache = compiled_cache
        self._conn_kw = kwargs

    @property
    def dialect(self):
        """An dialect for engine."""
        return self._dialect

    @property
    def name(self):
        """A name of the dialect."""
        return self._dialect.name

    @property
    def driver(self):
        """A driver of the dialect."""
        return self._dialect.driver

    @property
    def minsize(self):
        return self._pool.minsize

    @property
    def maxsize(self):
        return self._pool.maxsize

    @property
    def size(self):
        return self._pool.size

    @property
    def freesize(self):
        return self._pool.freesize

    def close(self):
        """Close engine.

        Mark all engine connections to be closed on getting back to pool.
        Closed engine doesn't allow to acquire new connections.
        """
        self._pool.close()

    def terminate(self):
        """Terminate engine.

        Terminate engine pool with instantly closing all acquired
        connections also.
        """
        self._pool.terminate()

    async def wait_closed(self):
        """Wait for closing all engine's connections."""
        await self._pool.wait_closed()

    def acquire(self):
        """Get a connection from pool."""
        coro = self._acquire()
        return _EngineAcquireContextManager(coro, self)

    async def _acquire(self):
        raw = await self._pool.acquire()
        conn = SAConnection(raw, self, compiled_cache=self._compiled_cache)
        return conn

    def release(self, conn):
        """Revert back connection to pool."""
        if conn.in_transaction:
            raise InvalidRequestError("Cannot release a connection with "
                                      "not finished transaction")
        raw = conn.connection
        return self._pool.release(raw)

    def __enter__(self):
        raise RuntimeError(
            '"yield from" should be used as context manager expression')

    def __exit__(self, *args):
        # This must exist because __enter__ exists, even though that
        # always raises; that's how the with-statement works.
        pass  # pragma: nocover

    def __iter__(self):
        # This is not a coroutine.  It is meant to enable the idiom:
        #
        #     with (yield from engine) as conn:
        #         <block>
        #
        # as an alternative to:
        #
        #     conn = yield from engine.acquire()
        #     try:
        #         <block>
        #     finally:
        #         engine.release(conn)
        conn = yield from self.acquire()
        return _ConnectionContextManager(self, conn)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.close()
        await self.wait_closed()


_EngineContextManager = _PoolContextManager
_EngineAcquireContextManager = _PoolAcquireContextManager


class _ConnectionContextManager:
    """Context manager.

    This enables the following idiom for acquiring and releasing a
    connection around a block:

        with (yield from engine) as conn:
            cur = yield from conn.cursor()

    while failing loudly when accidentally using:

        with engine:
            <block>
    """

    __slots__ = ('_engine', '_conn')

    def __init__(self, engine, conn):
        self._engine = engine
        self._conn = conn

    def __enter__(self):
        assert self._conn is not None
        return self._conn

    def __exit__(self, *args):
        try:
            self._engine.release(self._conn)
        finally:
            self._engine = None
            self._conn = None



import datetime
import json
import re
import time

import pytest
from pymysql import util
from pymysql.err import ProgrammingError


@pytest.fixture
def datatype_table(loop, cursor, table_cleanup):
    async def f():
        await cursor.execute(
            "CREATE TABLE test_datatypes (b bit, i int, l bigint, f real, s "
            "varchar(32), u varchar(32), bb blob, d date, dt datetime, "
            "ts timestamp, td time, t time, st datetime)")
        table_cleanup('test_datatypes')
    loop.run_until_complete(f())
    table_cleanup('test_datatypes')


@pytest.mark.run_loop
async def test_datatypes(connection, cursor, datatype_table):
    # insert values
    v = (
        True, -3, 123456789012, 5.7, "hello'\" world",
        u"Espa\xc3\xb1ol",
        "binary\x00data".encode(connection.charset),
        datetime.date(1988, 2, 2),
        datetime.datetime.now().replace(microsecond=0),
        datetime.timedelta(5, 6), datetime.time(16, 32),
        time.localtime())
    await cursor.execute(
        "INSERT INTO test_datatypes (b,i,l,f,s,u,bb,d,dt,td,t,st) "
        "values (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)",
        v)
    await cursor.execute(
        "select b,i,l,f,s,u,bb,d,dt,td,t,st from test_datatypes")
    r = await cursor.fetchone()
    assert util.int2byte(1) == r[0]
    # assert v[1:8] == r[1:8])
    assert v[1:9] == r[1:9]
    # mysql throws away microseconds so we need to check datetimes
    # specially. additionally times are turned into timedeltas.
    # self.assertEqual(datetime.datetime(*v[8].timetuple()[:6]), r[8])

    # TODO: figure out why this assert fails
    # assert [9] == r[9]  # just timedeltas
    expected = datetime.timedelta(0, 60 * (v[10].hour * 60 + v[10].minute))
    assert expected == r[10]
    assert datetime.datetime(*v[-1][:6]) == r[-1]


@pytest.mark.run_loop
async def test_datatypes_nulls(cursor, datatype_table):
    # check nulls
    await cursor.execute(
        "insert into test_datatypes (b,i,l,f,s,u,bb,d,dt,td,t,st) "
        "values (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)",
        [None] * 12)
    await cursor.execute(
        "select b,i,l,f,s,u,bb,d,dt,td,t,st from test_datatypes")
    r = await cursor.fetchone()
    assert tuple([None] * 12) == r


@pytest.mark.run_loop
async def test_datatypes_sequence_types(cursor, datatype_table):
    # check sequence type
    await cursor.execute(
        "INSERT INTO test_datatypes (i, l) VALUES (2,4), (6,8), "
        "(10,12)")
    await cursor.execute(
        "select l from test_datatypes where i in %s order by i",
        ((2, 6),))
    r = await cursor.fetchall()
    assert ((4,), (8,)) == r


@pytest.mark.run_loop
async def test_dict_escaping(cursor, table_cleanup):
    sql = "CREATE TABLE test_dict (a INTEGER, b INTEGER, c INTEGER)"
    await cursor.execute(sql)
    table_cleanup('test_dict')
    sql = "INSERT INTO test_dict (a,b,c) VALUES (%(a)s, %(b)s, %(c)s)"
    await cursor.execute(sql, {"a": 1, "b": 2, "c": 3})
    await cursor.execute("SELECT a,b,c FROM test_dict")
    r = await cursor.fetchone()
    assert (1, 2, 3) == r


@pytest.mark.run_loop
async def test_string(cursor, table_cleanup):
    await cursor.execute("DROP TABLE IF EXISTS test_string;")
    await cursor.execute("CREATE TABLE test_string (a text)")
    test_value = "I am a test string"
    table_cleanup('test_string')
    await cursor.execute("INSERT INTO test_string (a) VALUES (%s)",
                         test_value)
    await cursor.execute("SELECT a FROM test_string")
    r = await cursor.fetchone()
    assert (test_value,) == r


@pytest.mark.run_loop
async def test_integer(cursor, table_cleanup):
    await cursor.execute("CREATE TABLE test_integer (a INTEGER)")
    table_cleanup('test_integer')
    test_value = 12345
    await cursor.execute("INSERT INTO test_integer (a) VALUES (%s)",
                         test_value)
    await cursor.execute("SELECT a FROM test_integer")
    r = await cursor.fetchone()
    assert (test_value,) == r


@pytest.mark.run_loop
async def test_binary_data(cursor, table_cleanup):
    data = bytes(bytearray(range(256)) * 4)
    await cursor.execute("CREATE TABLE test_blob (b blob)")
    table_cleanup('test_blob')
    await cursor.execute("INSERT INTO test_blob (b) VALUES (%s)",
                         (data,))
    await cursor.execute("SELECT b FROM test_blob")
    (r,) = await cursor.fetchone()
    assert data == r


@pytest.mark.run_loop
async def test_untyped_convertion_to_null_and_empty_string(cursor):
    await cursor.execute("select null,''")
    r = await cursor.fetchone()
    assert (None, u'') == r
    await cursor.execute("select '',null")
    r = await cursor.fetchone()
    assert (u'', None) == r


@pytest.mark.run_loop
async def test_timedelta_conversion(cursor):
    await cursor.execute(
        "select time('12:30'), time('23:12:59'), time('23:12:59.05100'), "
        "time('-12:30'), time('-23:12:59'), time('-23:12:59.05100'), "
        "time('-00:30')")
    r = await cursor.fetchone()
    assert (datetime.timedelta(0, 45000),
            datetime.timedelta(0, 83579),
            datetime.timedelta(0, 83579, 51000),
            -datetime.timedelta(0, 45000),
            -datetime.timedelta(0, 83579),
            -datetime.timedelta(0, 83579, 51000),
            -datetime.timedelta(0, 1800)) == r


@pytest.mark.run_loop
async def test_datetime_conversion(cursor, table_cleanup):
    dt = datetime.datetime(2013, 11, 12, 9, 9, 9, 123450)
    try:
        await cursor.execute("CREATE TABLE test_datetime"
                             "(id INT, ts DATETIME(6))")
        table_cleanup('test_datetime')
        await cursor.execute("INSERT INTO test_datetime VALUES "
                             "(1,'2013-11-12 09:09:09.12345')")
        await cursor.execute("SELECT ts FROM test_datetime")
        r = await cursor.fetchone()
        assert (dt,) == r
    except ProgrammingError:
        # User is running a version of MySQL that doesn't support
        # msecs within datetime
        pass


@pytest.mark.run_loop
async def test_get_transaction_status(connection, cursor):
    #  make sure that connection is clean without transactions
    transaction_flag = connection.get_transaction_status()
    assert not transaction_flag

    # start transaction
    await connection.begin()
    # make sure transaction flag is up
    transaction_flag = connection.get_transaction_status()
    assert transaction_flag

    await cursor.execute('SELECT 1;')
    (r, ) = await cursor.fetchone()
    assert r == 1
    await connection.commit()
    # make sure that transaction flag is down
    transaction_flag = connection.get_transaction_status()
    assert not transaction_flag


@pytest.mark.run_loop
async def test_rollback(connection, cursor):
    await cursor.execute('DROP TABLE IF EXISTS tz_data;')
    await cursor.execute('CREATE TABLE tz_data ('
                         'region VARCHAR(64),'
                         'zone VARCHAR(64),'
                         'name VARCHAR(64))')
    await connection.commit()

    args = ('America', '', 'America/New_York')
    await cursor.execute('INSERT INTO tz_data VALUES (%s, %s, %s)',
                         args)
    await cursor.execute('SELECT * FROM tz_data;')
    data = await cursor.fetchall()
    assert len(data) == 1

    await connection.rollback()
    await cursor.execute('SELECT * FROM tz_data;')
    data = await cursor.fetchall()

    # should not return any rows since no inserts was commited
    assert len(data) == 0


def mysql_server_is(server_version, version_tuple):
    """Return True if the given connection is on the version given or
    greater.
    e.g.::
        if self.mysql_server_is(conn, (5, 6, 4)):
            # do something for MySQL 5.6.4 and above
    """
    server_version_tuple = tuple(
        (int(dig) if dig is not None else 0)
        for dig in
        re.match(r'(\d+)\.(\d+)\.(\d+)', server_version).group(1, 2, 3)
    )
    return server_version_tuple >= version_tuple


@pytest.mark.run_loop
async def test_json(connection_creator, table_cleanup):
    connection = await connection_creator(
        charset="utf8mb4", autocommit=True)
    server_info = connection.get_server_info()
    if not mysql_server_is(server_info, (5, 7, 0)):
        raise pytest.skip("JSON type is not supported on MySQL <= 5.6")

    cursor = await connection.cursor()
    await cursor.execute("""\
    CREATE TABLE test_json (
        id INT NOT NULL,
        json JSON NOT NULL,
        PRIMARY KEY (id)
    );""")
    table_cleanup("test_json")
    json_str = '{"hello": "こんにちは"}'
    await cursor.execute(
        "INSERT INTO test_json (id, `json`) values (42, %s)", (json_str,))
    await cursor.execute("SELECT `json` from `test_json` WHERE `id`=42")

    r = await cursor.fetchone()
    assert json.loads(r[0]) == json.loads(json_str)

    await cursor.execute("SELECT CAST(%s AS JSON) AS x", (json_str,))
    r = await cursor.fetchone()
    assert json.loads(r[0]) == json.loads(json_str)



import asyncio
import aiomysql
from aiomysql import connect, sa, Cursor

import os
import unittest
from unittest import mock

from sqlalchemy import MetaData, Table, Column, Integer, String
from sqlalchemy.schema import DropTable, CreateTable


meta = MetaData()
tbl = Table('sa_tbl', meta,
            Column('id', Integer, nullable=False,
                   primary_key=True),
            Column('name', String(255)))


class TestSAConnection(unittest.TestCase):
    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        self.host = os.environ.get('MYSQL_HOST', 'localhost')
        self.port = int(os.environ.get('MYSQL_PORT', 3306))
        self.user = os.environ.get('MYSQL_USER', 'root')
        self.db = os.environ.get('MYSQL_DB', 'test_pymysql')
        self.password = os.environ.get('MYSQL_PASSWORD', '')

    def tearDown(self):
        self.loop.close()

    async def connect(self, **kwargs):
        conn = await connect(db=self.db,
                             user=self.user,
                             password=self.password,
                             host=self.host,
                             loop=self.loop,
                             port=self.port,
                             **kwargs)
        await conn.autocommit(True)
        cur = await conn.cursor()
        await cur.execute("DROP TABLE IF EXISTS sa_tbl")
        await cur.execute("CREATE TABLE sa_tbl "
                          "(id serial, name varchar(255))")
        await cur.execute("INSERT INTO sa_tbl (name)"
                          "VALUES ('first')")

        await cur._connection.commit()
        # yield from cur.close()
        engine = mock.Mock()
        engine.dialect = sa.engine._dialect
        return sa.SAConnection(conn, engine)

    def test_execute_text_select(self):
        async def go():
            conn = await self.connect()
            res = await conn.execute("SELECT * FROM sa_tbl;")
            self.assertIsInstance(res.cursor, Cursor)
            self.assertEqual(('id', 'name'), res.keys())
            rows = await res.fetchall()
            self.assertTrue(res.closed)
            self.assertIsNone(res.cursor)
            self.assertEqual(1, len(rows))
            row = rows[0]
            self.assertEqual(1, row[0])
            self.assertEqual(1, row['id'])
            self.assertEqual(1, row.id)
            self.assertEqual('first', row[1])
            self.assertEqual('first', row['name'])
            self.assertEqual('first', row.name)
            # TODO: fix this
            await conn._connection.commit()
        self.loop.run_until_complete(go())

    def test_execute_sa_select(self):
        async def go():
            conn = await self.connect()
            res = await conn.execute(tbl.select())
            self.assertIsInstance(res.cursor, Cursor)
            self.assertEqual(('id', 'name'), res.keys())
            rows = await res.fetchall()
            self.assertTrue(res.closed)
            self.assertIsNone(res.cursor)
            self.assertTrue(res.returns_rows)

            self.assertEqual(1, len(rows))
            row = rows[0]
            self.assertEqual(1, row[0])
            self.assertEqual(1, row['id'])
            self.assertEqual(1, row.id)
            self.assertEqual('first', row[1])
            self.assertEqual('first', row['name'])
            self.assertEqual('first', row.name)
            # TODO: fix this
            await conn._connection.commit()

        self.loop.run_until_complete(go())

    def test_execute_sa_insert_with_dict(self):
        async def go():
            conn = await self.connect()
            await conn.execute(tbl.insert(), {"id": 2, "name": "second"})

            res = await conn.execute(tbl.select())
            rows = await res.fetchall()
            self.assertEqual(2, len(rows))
            self.assertEqual((1, 'first'), rows[0])
            self.assertEqual((2, 'second'), rows[1])

        self.loop.run_until_complete(go())

    def test_execute_sa_insert_with_tuple(self):
        async def go():
            conn = await self.connect()
            await conn.execute(tbl.insert(), (2, "second"))

            res = await conn.execute(tbl.select())
            rows = await res.fetchall()
            self.assertEqual(2, len(rows))
            self.assertEqual((1, 'first'), rows[0])
            self.assertEqual((2, 'second'), rows[1])

        self.loop.run_until_complete(go())

    def test_execute_sa_insert_named_params(self):
        async def go():
            conn = await self.connect()
            await conn.execute(tbl.insert(), id=2, name="second")

            res = await conn.execute(tbl.select())
            rows = await res.fetchall()
            self.assertEqual(2, len(rows))
            self.assertEqual((1, 'first'), rows[0])
            self.assertEqual((2, 'second'), rows[1])

        self.loop.run_until_complete(go())

    def test_execute_sa_insert_positional_params(self):
        async def go():
            conn = await self.connect()
            await conn.execute(tbl.insert(), 2, "second")

            res = await conn.execute(tbl.select())
            rows = await res.fetchall()
            self.assertEqual(2, len(rows))
            self.assertEqual((1, 'first'), rows[0])
            self.assertEqual((2, 'second'), rows[1])

        self.loop.run_until_complete(go())

    def test_scalar(self):
        async def go():
            conn = await self.connect()
            res = await conn.scalar(tbl.count())
            self.assertEqual(1, res)

        self.loop.run_until_complete(go())

    def test_scalar_None(self):
        async def go():
            conn = await self.connect()
            await conn.execute(tbl.delete())
            res = await conn.scalar(tbl.select())
            self.assertIsNone(res)
            # TODO: fix this
            await conn._connection.commit()

        self.loop.run_until_complete(go())

    def test_row_proxy(self):
        async def go():
            conn = await self.connect()
            res = await conn.execute(tbl.select())
            rows = await res.fetchall()
            row = rows[0]
            row2 = await (await conn.execute(tbl.select())).first()
            self.assertEqual(2, len(row))
            self.assertEqual(['id', 'name'], list(row))
            self.assertIn('id', row)
            self.assertNotIn('unknown', row)
            self.assertEqual('first', row.name)
            self.assertEqual('first', row[tbl.c.name])
            with self.assertRaises(AttributeError):
                row.unknown
            self.assertEqual("(1, 'first')", repr(row))
            self.assertEqual((1, 'first'), row.as_tuple())
            self.assertNotEqual((555, 'other'), row.as_tuple())
            self.assertEqual(row2, row)
            self.assertFalse(row2 != row)
            self.assertNotEqual(5, row)
            # TODO: fix this
            await conn._connection.commit()

        self.loop.run_until_complete(go())

    def test_insert(self):
        async def go():
            conn = await self.connect()
            res = await conn.execute(tbl.insert().values(name='second'))
            self.assertEqual(1, res.rowcount)
            self.assertEqual(2, res.lastrowid)

        self.loop.run_until_complete(go())

    def test_raw_insert(self):
        async def go():
            conn = await self.connect()
            await conn.execute(
                "INSERT INTO sa_tbl (name) VALUES ('third')")
            res = await conn.execute(tbl.select())
            self.assertEqual(2, res.rowcount)
            self.assertEqual(('id', 'name'), res.keys())
            self.assertTrue(res.returns_rows)

            rows = await res.fetchall()
            self.assertEqual(2, len(rows))
            self.assertEqual(2, rows[1].id)
        self.loop.run_until_complete(go())

    def test_raw_insert_with_params(self):
        async def go():
            conn = await self.connect()
            res = await conn.execute(
                "INSERT INTO sa_tbl (id, name) VALUES (%s, %s)",
                2, 'third')
            res = await conn.execute(tbl.select())
            self.assertEqual(2, res.rowcount)
            self.assertEqual(('id', 'name'), res.keys())
            self.assertTrue(res.returns_rows)

            rows = await res.fetchall()
            self.assertEqual(2, len(rows))
            self.assertEqual(2, rows[1].id)
        self.loop.run_until_complete(go())

    def test_raw_insert_with_params_dict(self):
        async def go():
            conn = await self.connect()
            res = await conn.execute(
                "INSERT INTO sa_tbl (id, name) VALUES (%(id)s, %(name)s)",
                {'id': 2, 'name': 'third'})
            res = await conn.execute(tbl.select())
            self.assertEqual(2, res.rowcount)
            self.assertEqual(('id', 'name'), res.keys())
            self.assertTrue(res.returns_rows)

            rows = await res.fetchall()
            self.assertEqual(2, len(rows))
            self.assertEqual(2, rows[1].id)
        self.loop.run_until_complete(go())

    def test_raw_insert_with_named_params(self):
        async def go():
            conn = await self.connect()
            res = await conn.execute(
                "INSERT INTO sa_tbl (id, name) VALUES (%(id)s, %(name)s)",
                id=2, name='third')
            res = await conn.execute(tbl.select())
            self.assertEqual(2, res.rowcount)
            self.assertEqual(('id', 'name'), res.keys())
            self.assertTrue(res.returns_rows)

            rows = await res.fetchall()
            self.assertEqual(2, len(rows))
            self.assertEqual(2, rows[1].id)
        self.loop.run_until_complete(go())

    def test_raw_insert_with_executemany(self):
        async def go():
            conn = await self.connect()
            with self.assertRaises(sa.ArgumentError):
                await conn.execute(
                    "INSERT INTO sa_tbl (id, name) VALUES (%(id)s, %(name)s)",
                    [(2, 'third'), (3, 'forth')])
        self.loop.run_until_complete(go())

    def test_raw_select_with_wildcard(self):
        async def go():
            conn = await self.connect()
            await conn.execute(
                'SELECT * FROM sa_tbl WHERE name LIKE "%test%"')
        self.loop.run_until_complete(go())

    def test_delete(self):
        async def go():
            conn = await self.connect()

            res = await conn.execute(tbl.delete().where(tbl.c.id == 1))

            self.assertEqual((), res.keys())
            self.assertEqual(1, res.rowcount)
            self.assertFalse(res.returns_rows)
            self.assertTrue(res.closed)
            self.assertIsNone(res.cursor)

        self.loop.run_until_complete(go())

    def test_double_close(self):
        async def go():
            conn = await self.connect()
            res = await conn.execute("SELECT 1")
            await res.close()
            self.assertTrue(res.closed)
            self.assertIsNone(res.cursor)
            await res.close()
            self.assertTrue(res.closed)
            self.assertIsNone(res.cursor)

        self.loop.run_until_complete(go())

    @unittest.skip("Find out how to close cursor on __del__ method")
    def test_weakrefs(self):
        async def go():
            conn = await self.connect()
            self.assertEqual(0, len(conn._weak_results))
            res = await conn.execute("SELECT 1")
            self.assertEqual(1, len(conn._weak_results))
            cur = res.cursor
            self.assertFalse(cur.closed)
            # TODO: fix this, how close cursor if result was deleted
            # yield from cur.close()
            del res
            self.assertTrue(cur.closed)
            self.assertEqual(0, len(conn._weak_results))

        self.loop.run_until_complete(go())

    def test_fetchall(self):
        async def go():
            conn = await self.connect()
            await conn.execute(tbl.insert().values(name='second'))

            res = await conn.execute(tbl.select())
            rows = await res.fetchall()
            self.assertEqual(2, len(rows))
            self.assertTrue(res.closed)
            self.assertTrue(res.returns_rows)
            self.assertEqual([(1, 'first'), (2, 'second')], rows)

        self.loop.run_until_complete(go())

    def test_fetchall_closed(self):
        async def go():
            conn = await self.connect()
            await conn.execute(tbl.insert().values(name='second'))

            res = await conn.execute(tbl.select())
            await res.close()
            with self.assertRaises(sa.ResourceClosedError):
                await res.fetchall()

        self.loop.run_until_complete(go())

    def test_fetchall_not_returns_rows(self):
        async def go():
            conn = await self.connect()
            res = await conn.execute(tbl.delete())
            with self.assertRaises(sa.ResourceClosedError):
                await res.fetchall()

        self.loop.run_until_complete(go())

    def test_fetchone_closed(self):
        async def go():
            conn = await self.connect()
            await conn.execute(tbl.insert().values(name='second'))

            res = await conn.execute(tbl.select())
            await res.close()
            with self.assertRaises(sa.ResourceClosedError):
                await res.fetchone()

        self.loop.run_until_complete(go())

    def test_first_not_returns_rows(self):
        async def go():
            conn = await self.connect()
            res = await conn.execute(tbl.delete())
            with self.assertRaises(sa.ResourceClosedError):
                await res.first()

        self.loop.run_until_complete(go())

    def test_fetchmany(self):
        async def go():
            conn = await self.connect()
            await conn.execute(tbl.insert().values(name='second'))

            res = await conn.execute(tbl.select())
            rows = await res.fetchmany()
            self.assertEqual(1, len(rows))
            self.assertFalse(res.closed)
            self.assertTrue(res.returns_rows)
            self.assertEqual([(1, 'first')], rows)

        self.loop.run_until_complete(go())

    def test_fetchmany_with_size(self):
        async def go():
            conn = await self.connect()
            await conn.execute(tbl.insert().values(name='second'))

            res = await conn.execute(tbl.select())
            rows = await res.fetchmany(100)
            self.assertEqual(2, len(rows))
            self.assertFalse(res.closed)
            self.assertTrue(res.returns_rows)
            self.assertEqual([(1, 'first'), (2, 'second')], rows)

        self.loop.run_until_complete(go())

    def test_fetchmany_closed(self):
        async def go():
            conn = await self.connect()
            await conn.execute(tbl.insert().values(name='second'))

            res = await conn.execute(tbl.select())
            await res.close()
            with self.assertRaises(sa.ResourceClosedError):
                await res.fetchmany()

        self.loop.run_until_complete(go())

    def test_fetchmany_with_size_closed(self):
        async def go():
            conn = await self.connect()
            await conn.execute(tbl.insert().values(name='second'))

            res = await conn.execute(tbl.select())
            await res.close()
            with self.assertRaises(sa.ResourceClosedError):
                await res.fetchmany(5555)

        self.loop.run_until_complete(go())

    def test_fetchmany_not_returns_rows(self):
        async def go():
            conn = await self.connect()
            res = await conn.execute(tbl.delete())
            with self.assertRaises(sa.ResourceClosedError):
                await res.fetchmany()

        self.loop.run_until_complete(go())

    def test_fetchmany_close_after_last_read(self):
        async def go():
            conn = await self.connect()

            res = await conn.execute(tbl.select())
            rows = await res.fetchmany()
            self.assertEqual(1, len(rows))
            self.assertFalse(res.closed)
            self.assertTrue(res.returns_rows)
            self.assertEqual([(1, 'first')], rows)
            rows2 = await res.fetchmany()
            self.assertEqual(0, len(rows2))
            self.assertTrue(res.closed)

        self.loop.run_until_complete(go())

    def test_create_table(self, **kwargs):
        async def go():
            conn = await self.connect()
            res = await conn.execute(DropTable(tbl))
            with self.assertRaises(sa.ResourceClosedError):
                await res.fetchmany()

            with self.assertRaises(aiomysql.ProgrammingError):
                await conn.execute("SELECT * FROM sa_tbl")

            res = await conn.execute(CreateTable(tbl))
            with self.assertRaises(sa.ResourceClosedError):
                await res.fetchmany()

            res = await conn.execute("SELECT * FROM sa_tbl")
            self.assertEqual(0, len(await res.fetchall()))

        self.loop.run_until_complete(go())



import unittest
import sqlalchemy

from aiomysql.sa.connection import _distill_params


class DistillArgsTest(unittest.TestCase):
    def test_distill_none(self):
        self.assertEqual(
            _distill_params(None, None),
            []
        )

    def test_distill_no_multi_no_param(self):
        self.assertEqual(
            _distill_params((), {}),
            []
        )

    def test_distill_dict_multi_none_param(self):
        self.assertEqual(
            _distill_params(None, {"foo": "bar"}),
            [{"foo": "bar"}]
        )

    def test_distill_dict_multi_empty_param(self):
        self.assertEqual(
            _distill_params((), {"foo": "bar"}),
            [{"foo": "bar"}]
        )

    def test_distill_single_dict(self):
        self.assertEqual(
            _distill_params(({"foo": "bar"},), {}),
            [{"foo": "bar"}]
        )

    def test_distill_single_list_strings(self):
        self.assertEqual(
            _distill_params((["foo", "bar"],), {}),
            [["foo", "bar"]]
        )

    def test_distill_single_list_tuples(self):
        self.assertEqual(
            _distill_params(([("foo", "bar"), ("bat", "hoho")],), {}),
            [('foo', 'bar'), ('bat', 'hoho')]
        )

    def test_distill_single_list_tuple(self):
        self.assertEqual(
            _distill_params(([("foo", "bar")],), {}),
            [('foo', 'bar')]
        )

    def test_distill_multi_list_tuple(self):
        self.assertEqual(
            _distill_params(
                ([("foo", "bar")], [("bar", "bat")]),
                {}
            ),
            ([('foo', 'bar')], [('bar', 'bat')])
        )

    def test_distill_multi_strings(self):
        self.assertEqual(
            _distill_params(("foo", "bar"), {}),
            [('foo', 'bar')]
        )

    def test_distill_single_list_dicts(self):
        self.assertEqual(
            _distill_params(([{"foo": "bar"}, {"foo": "hoho"}],), {}),
            [{'foo': 'bar'}, {'foo': 'hoho'}]
        )

    def test_distill_single_string(self):
        self.assertEqual(
            _distill_params(("arg",), {}),
            [["arg"]]
        )

    def test_distill_multi_string_tuple(self):
        self.assertEqual(
            _distill_params((("arg", "arg"),), {}),
            [("arg", "arg")]
        )


sqlalchemy  # for sake of pyflakes checks
